/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : USE_DMY_CLOCK, PRC_WATCH, GPR_INITCLR

//`define USE_DMY_CLOCK
//`define PRC_WATCH
//`define GPR_INITCLR

// local defines : 

// GPR modele for the Reduced Core TinyAVR (attiny10) 
// version 005
// 
// Target Device: xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal: Balanced
//                                      
//                                       TUNE_3   TUNE_2
// Number of Slice Flip Flops:              56      64
// Number of 4 input LUTs:                 145     146
// Number of occupied Slices:               93      94
// Total Number of 4 input LUTs:           148     149
// Number used for Dual Port RAMs:          24      32
// Number of bonded IOBs:                   79      79
// Post-PAR Static Timing Report
// Maximum frequency:                 68.681MHz  70.611MHz
//
module rtavr_gpr_16(
		input RESET,
`ifdef USE_DMY_CLOCK
		input CLK2X_IN,
		output CLK_OUT,
`else
		input CLK2X,
		input CLK,
`endif
		input WE,
		input PREDEC, POSTINC, 
		input [4:0] IDX_OFF,
		input [7:0] DI,
		input [3:0] ADDRAL,
		input [3:0] ADDRAH,
		input [3:0] ADDRBL,
		input [3:0] ADDRBH,
		output [7:0] DOBL,
		output [7:0] DOBH
		, output [11:0] JA
		, output [15:0] INDEX
  `ifdef FLAGS_TUNE_2
		, output EQ_BL_BH
  `endif
  `ifdef SBRX_TUNE
		, input[2:0] BL_BIT_SEL
		, output BL_BIT_OUT
  `endif
`ifdef PRC_WATCH
		, output [3:0] WB_ADDR
`endif
		);
// Timing MEMO:
//
//                T(N-1)               |            TN                 |
//              _______         _______         _______         _______
// CLK2X ______|       |_______|       |_______|       |_______|       |
//                       _______________                 _______________
// CLK _________________|               |_______________|               |
//
// ADDRAL,      ================================ (PM_DRIVEN_CLK2X) 
// ADDRBL                ================================ (else)     
// ADDRAH,ADDRBH,                       =================================
// PREDEC,POSTINC,WE
// DI                                                   ================
//                                     V                V      V
//                                 ADDRAL            ADDRAH   DI
//                                 ADDRBL            ADDRBH
//                   
// DOBL,DOBL                                                            |===
// INDEX(lo8)                                  ===========================..
// INDEX(hi8)                                           ==================
// JA(lo8)                                                      FFFFFFFFFF..
// JA(hi8)                                                      FFFFFFFFFF..

`ifdef USE_DMY_CLOCK
  reg CLK;
  wire CLK2X;
  assign CLK_OUT = CLK;
  assign CLK2X   = CLK2X_IN;
`endif
  reg  [3:0] r_addra;
  reg  [3:0] r_addrb;

  reg  [7:0] s2_DOBL;
  reg  [7:0] s2_DOBH;
  reg  [7:0] gpr [0:15];

  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;

  reg  c;
  reg  ci;
  wire WB = (PREDEC | POSTINC);
  reg  [3:0] r_wb_addr;

  `ifdef PRC_WATCH
    assign WB_ADDR = r_wb_addr;
  `endif

  wire [7:0] DOA = gpr[r_addra];
  wire [7:0] DOB = gpr[r_addrb];

  wire [7:0] DOA2LP = DOA - 1;
  wire [8:0] DOA2LO = {1'b0 , DOA} + { 4'b0000 ,IDX_OFF };
  wire [7:0] DOA2H = (PREDEC &  c) ? DOB - 1 : DOB ;

  wire [8:0] WB_DATA = (PREDEC & (~CLK | c) ) ?  {1'b0 , DOA} - 1
                     : (POSTINC & (~CLK | c)) ?  {1'b0 , DOA} + 1
		     : {1'b0 , DOA};
  
  wire WE2 = WB | (WE & CLK) ;

  reg  [7:0] s1_DOAL;
  reg  [7:0] s1_DOBL;

  reg r_forward_to_bl;
  wire FORWARD_TO_BL = r_forward_to_bl;

  wire [7:0] v_DOBL = FORWARD_TO_BL ? DI : s1_DOBL;

  wire [7:0] v_DOBH = gpr[r_addrb];

    assign JA[7:0] = FORWARD_TO_BL ? DI : s1_DOBL;
    assign INDEX[7:0] = s1_DOAL;

    assign JA[11:8] = gpr[r_addrb][3:0];
	
    assign INDEX[15:13] = DOA2H[7:5];
    assign INDEX[12:8] =  DOA2H[4:0]
			+ ( ci ? 1 : 0 );
  `ifdef FLAGS_TUNE_2
	reg s2_eq_bl_bh;
	assign EQ_BL_BH = s2_eq_bl_bh;
  `endif
  `ifdef SBRX_TUNE
	reg s2_bl_bit_out;
	assign BL_BIT_OUT = s2_bl_bit_out;

	wire v_bl_bit_out =  (BL_BIT_SEL == 0) ? v_DOBL[0]
			   : (BL_BIT_SEL == 1) ? v_DOBL[1] 
			   : (BL_BIT_SEL == 2) ? v_DOBL[2]
			   : (BL_BIT_SEL == 3) ? v_DOBL[3]
			   : (BL_BIT_SEL == 4) ? v_DOBL[4]
			   : (BL_BIT_SEL == 5) ? v_DOBL[5]
			   : (BL_BIT_SEL == 6) ? v_DOBL[6]
			   : v_DOBL[7];

  `endif
`ifdef USE_DMY_CLOCK
always @(negedge CLK2X)
  begin
	CLK <= ~CLK;
  end

initial
  begin
	CLK <= 0;
  end
`endif

always @(posedge CLK2X)
  begin
    if (RESET == 0)
      begin
	s1_DOAL <= 0;
	r_wb_addr <= 0;
      end
    else
      begin
	if (WE2) gpr[r_addra] <= (WB) ? WB_DATA : DI ;
	if (~CLK) s1_DOAL <= (PREDEC) ? DOA2LP[7:0] : DOA2LO[7:0] ;
	if (~CLK) r_wb_addr <= (WE) ? ADDRAH : ADDRBH;
      end
  end

always @(negedge CLK2X)
  begin
    if (RESET == 0)
      begin
	r_addra <= 0;
	r_addrb <= 0;
	s1_DOBL <= 0;
	s2_DOBL <= 0;
	s2_DOBH <= 0;
  `ifdef FLAGS_TUNE_2
	s2_eq_bl_bh <= 0;
  `endif
  `ifdef SBRX_TUNE
	s2_bl_bit_out <= 0;
  `endif
	r_forward_to_bl <= 0;
	c <= 0;
	ci <= 0;
      end
    else
      begin
	r_addra <= CLK ? ADDRAL : r_wb_addr;
	r_addrb <= CLK ? ADDRBL : ADDRBH;
	if (~CLK) s1_DOBL <= gpr[r_addrb];
	if (CLK) s2_DOBL <= v_DOBL;
	if (CLK) s2_DOBH <= v_DOBH;
  `ifdef FLAGS_TUNE_2
	if (CLK) s2_eq_bl_bh <= ( v_DOBL == v_DOBH);
  `endif
  `ifdef SBRX_TUNE
	if (CLK) s2_bl_bit_out <= v_bl_bit_out;
  `endif

//	if (~CLK) r_forward_to_bl <=  (WE) & (r_wb_addr == ADDRBL);
	if (~CLK) r_forward_to_bl <=  (WE) & (r_wb_addr == r_addrb);
	c <= WB_DATA[8];
	ci <= DOA2LO[8];
      end
  end

`ifdef GPR_INITCLR
  integer i;
  initial
    begin
      for (i = 0 ; i < 16 ; i = i + 1)
	gpr[i] <= 8'h00;
    end
`endif
endmodule
